#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File              : waves.py
# License           : MIT license <Check LICENSE>
# Author            : Anderson I. da Silva (aignacio) <anderson@aignacio.com>
# Date              : 25.10.2024
# Last Modified Date: 26.10.2024
import cocotb
import wavedrom
import json
import copy
import logging
import datetime

from .version import __version__
from cocotb.handle import SimHandleBase
from cocotb.triggers import RisingEdge, FallingEdge


class signal_data:
    def __init__(
        self,
        name,
        handle,
        is_clock,
        is_posedge_clock,
        color_data=None,
        previous_val=None,
        clock_period=1,
    ) -> None:
        self.name = name
        self.handle = handle
        self.color_data = color_data
        self.is_clock = is_clock
        self.is_posedge_clock = is_posedge_clock
        self.previous_val = previous_val
        self.clock_period = clock_period


class waveform:
    def __init__(
        self,
        clk,
        name,
        hscale: int = 2,
        is_posedge: bool = True,
        debug: bool = False,
    ) -> None:
        self.handles = []  # List to store [signal_data] obj
        self.waves = {}
        self.waves["signal"] = []
        self.trigger = {"trigger": None, "value": 0}

        self.head = {"text": name, "tick": 0, "every": 1}
        self.foot = {"text": "Generated by cocotbext-waves", "tick": 0, "every": 1}

        self.hscale = hscale
        self.debug = debug
        self.is_posedge = is_posedge
        self.clk = clk
        self.name = name
        self.color_idx = 3  # Start color for multi-bit signal

        self.add_signal(clk, color=None, is_clock=True, is_posedge_clock=is_posedge)
        self.mon = cocotb.start_soon(self._monitor())

        self.log = logging.getLogger(f"cocotb.waves.{name}")
        self.log.info(f"Waveform / Wavedrom - ({name})")
        self.log.info("cocotbext-waves version %s", __version__)
        self.log.info(
            f"Copyright (c) {datetime.datetime.now().year} Anderson Ignacio da Silva"
        )
        self.log.info("https://github.com/aignacio/cocotbext-waves")

    def add_trigger(self, handle, val):
        self.trigger = {"trigger": handle, "value": val}

    def add_signal(
        self,
        sig,
        color=None,
        is_clock: bool = False,
        is_posedge_clock: bool = True,
        clock_period=1,
    ):
        if not isinstance(sig, list):
            sig = [sig]

        for signal in sig:
            if signal.__len__() == 1:
                self.waves["signal"].append(
                    {
                        "name": signal._name,
                        "wave": "",
                    }
                )
                self.handles.append(
                    signal_data(
                        name=signal._name,
                        handle=signal,
                        is_clock=is_clock,
                        is_posedge_clock=is_posedge_clock,
                        clock_period=clock_period,
                    )
                )
            else:
                self.waves["signal"].append(
                    {
                        "name": signal._name,
                        "wave": "",
                        "data": "",
                    }
                )

                if color is None:
                    # Automatically assigns a color
                    if self.color_idx == 9:
                        color_tmp = 3
                    else:
                        color_tmp = self.color_idx
                        self.color_idx = color_tmp + 1
                else:
                    # Take user's input
                    color_tmp = color

                # Append to the list that will be used later
                self.handles.append(
                    signal_data(
                        name=signal._name,
                        handle=signal,
                        color_data=str(color_tmp),
                        is_clock=is_clock,
                        is_posedge_clock=is_posedge_clock,
                        clock_period=clock_period,
                    )
                )

    async def _monitor(self):
        while True:
            if self.is_posedge is True:
                await RisingEdge(self.clk)
                await FallingEdge(self.clk)
            else:
                await FallingEdge(self.clk)
                await RisingEdge(self.clk)

            if self.trigger["trigger"] is not None:
                if self.trigger["trigger"].value == self.trigger["value"]:
                    self._add_signals()
            else:
                self._add_signals()

    def _add_signals(self):
        for index, signal in enumerate(self.handles):
            self._append_wave_dot(index, signal)

    def _append_wave_dot(self, index, signal):
        if signal.is_clock is True:
            for entry in self.waves["signal"]:
                if entry["name"] == signal.name:
                    if signal.handle.value.is_resolvable is not True:
                        if "z" in signal.handle.value:
                            entry["wave"] += "z"
                        else:
                            entry["wave"] += "x"
                    else:
                        if entry["wave"] == "":
                            entry["wave"] += "P" if signal.is_posedge_clock else "N"
                            entry["period"] = signal.clock_period
                        else:
                            entry["wave"] += "."
        else:
            for entry in self.waves["signal"]:
                if entry["name"] == signal.name:
                    if signal.previous_val is None:
                        self.handles[index].previous_val = copy.deepcopy(
                            signal.handle.value
                        )

                        if signal.handle.value.is_resolvable is not True:
                            if "z" in signal.handle.value:
                                entry["wave"] += "z"
                            else:
                                entry["wave"] += "x"
                        else:
                            if signal.handle.__len__() > 1:
                                entry["data"] += str(hex(signal.handle.value)) + " "
                                entry["wave"] += signal.color_data
                            else:
                                entry["wave"] += str(signal.handle.value)
                    elif signal.handle.value == signal.previous_val:
                        entry["wave"] += "."
                    else:
                        self.handles[index].previous_val = copy.deepcopy(
                            signal.handle.value
                        )

                        if signal.handle.value.is_resolvable is not True:
                            if "z" in signal.handle.value:
                                entry["wave"] += "z"
                            else:
                                entry["wave"] += "x"
                        else:
                            if signal.handle.__len__() > 1:
                                entry["wave"] += signal.color_data
                                entry["data"] += str(hex(signal.handle.value)) + " "
                            else:
                                entry["wave"] += str(signal.handle.value)

    def set_head(self, text, tick, every):
        self.head = {"text": text, "tick": tick, "every": every}

    def set_foot(self, text, tick, every):
        self.foot = {"text": text, "tick": tick, "every": every}

    def save(self):
        self.mon.kill()
        # Format each name entry
        for signal in self.handles:
            for entry in self.waves["signal"]:
                if entry["name"] == signal.handle._name:
                    if signal.handle.__len__() > 1:
                        entry["name"] += "[" + str(signal.handle.__len__() - 1) + ":0]"

        self.waves["config"] = {"hscale": self.hscale}
        self.waves["head"] = self.head
        self.waves["foot"] = self.foot

        if self.debug:
            print("[Waves - Debug] Printing JSON Wavedrom")
            print(json.dumps(self.waves))
        svg = wavedrom.render(json.dumps(self.waves))
        svg.saveas(self.name + ".svg")

    def save_txt(self):
        try:
            with open(self.name + ".txt", "w") as file:
                json.dump(self.waves, file, indent=4)  # indent for better readability
            if self.debug:
                print(f"Wavedrom diagram written into {self.name}.txt")
        except Exception as e:
            print(f"An error occurred: {e} while trying to write wavedrom diagram")
